import copy

import pytest

import abjad


def test_Container___contains___01():
    note = abjad.Note("c'4")
    voice = abjad.Voice([abjad.Note("c'4")])

    assert note not in voice


def test_Container___copy___01():
    """
    Containers copy simultaneity flag.
    """

    container_1 = abjad.Container([abjad.Voice("c'8 d'8"), abjad.Voice("c''8 b'8")])
    container_1.simultaneous = True
    container_2 = copy.copy(container_1)

    assert abjad.lilypond(container_1) == abjad.string.normalize(
        r"""
        <<
            \new Voice
            {
                c'8
                d'8
            }
            \new Voice
            {
                c''8
                b'8
            }
        >>
        """
    )

    assert abjad.lilypond(container_2) == abjad.string.normalize(
        r"""
        <<
        >>
        """
    )

    assert container_1 is not container_2


def test_Container___delitem___01():
    """
    Deletes in-score container.
    """

    voice = abjad.Voice("{ c'8 ( d'8 ) } { e'8 ( f'8 ) }")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                (
                d'8
                )
            }
            {
                e'8
                (
                f'8
                )
            }
        }
        """
    )

    container = voice[0]
    del voice[0]

    # container no longer appears in score
    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                e'8
                (
                f'8
                )
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)

    # container leaves are still slurred
    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            (
            d'8
            )
        }
        """
    ), print(abjad.lilypond(container))

    assert abjad.wf.wellformed(container)


def test_Container___delitem___02():
    """
    Deletes in-score leaf.
    """

    voice = abjad.Voice("c'8 [ d'8 e'8 f'8 ]")
    del voice[1]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            e'8
            f'8
            ]
        }
        """
    )

    assert abjad.wf.wellformed(voice)


def test_Container___delitem___03():
    """
    Deletes slice in middle of container.
    """

    voice = abjad.Voice("c'8 [ d'8 e'8 f'8 ]")
    del voice[1:3]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            f'8
            ]
        }
        """
    )

    assert abjad.wf.wellformed(voice)


def test_Container___delitem___04():
    """
    Delete slice at beginning of container.
    """

    voice = abjad.Voice("c'8 [ d'8 e'8 f'8 ]")
    del voice[:2]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            e'8
            f'8
            ]
        }
        """
    )

    assert abjad.wf.wellformed(voice, check_overlapping_beams=False)


def test_Container___delitem___05():
    """
    Deletes slice at end of container.
    """

    voice = abjad.Voice("c'8 [ d'8 e'8 f'8 ]")
    del voice[2:]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
        }
        """
    )

    assert abjad.wf.wellformed(voice)


def test_Container___delitem___06():
    """
    Deletes container contents.
    """

    voice = abjad.Voice("c'8 [ d'8 e'8 f'8 ]")
    del voice[:]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
        }
        """
    )

    assert not len(voice)


def test_Container___delitem___07():
    """
    Deletes leaf from tuplet.
    """

    tuplet = abjad.Tuplet((2, 3), "c'8 [ d'8 e'8 ]")
    del tuplet[1]

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tweak edge-height #'(0.7 . 0)
        \tuplet 3/2
        {
            c'8
            [
            e'8
            ]
        }
        """
    )

    assert abjad.wf.wellformed(tuplet)


def test_Container___delitem___08():
    """
    Deletes leaf from nested container.
    """

    voice = abjad.Voice("c'8 [ { d'8 e'8 } f'8 ]")
    leaves = abjad.select.leaves(voice)
    abjad.glissando(leaves)

    string = abjad.lilypond(voice)
    assert string == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            \glissando
            {
                d'8
                \glissando
                e'8
                \glissando
            }
            f'8
            ]
        }
        """
    ), print(string)

    leaf = leaves[1]
    del voice[1][0]

    string = abjad.lilypond(voice)
    assert string == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            \glissando
            {
                e'8
                \glissando
            }
            f'8
            ]
        }
        """
    ), print(string)

    assert abjad.wf.wellformed(voice)
    assert abjad.wf.wellformed(leaf)


def test_Container___delitem___09():
    """
    REGRESSION. Container deletion does not orphan dependent wrappers.
    """

    voice = abjad.Voice("{ c'4 [ } d'4")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'4
                [
            }
            d'4
        }
        """
    )

    assert len(voice._dependent_wrappers) == 1

    del voice[0]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            d'4
        }
        """
    ), print(abjad.lilypond(voice))

    assert len(voice._dependent_wrappers) == 0

    assert abjad.wf.wellformed(voice)


def test_Container___getitem___01():
    """
    Get one container component with positive index.
    """

    notes = [
        abjad.Note("c'8"),
        abjad.Note("d'8"),
        abjad.Note("e'8"),
        abjad.Note("f'8"),
    ]
    voice = abjad.Voice(notes)

    assert voice[0] is notes[0]
    assert voice[1] is notes[1]
    assert voice[2] is notes[2]
    assert voice[3] is notes[3]


def test_Container___getitem___02():
    """
    Get one container component with negative index.
    """

    notes = [
        abjad.Note("c'8"),
        abjad.Note("d'8"),
        abjad.Note("e'8"),
        abjad.Note("f'8"),
    ]
    voice = abjad.Voice(notes)

    assert voice[-1] is notes[3]
    assert voice[-2] is notes[2]
    assert voice[-3] is notes[1]
    assert voice[-4] is notes[0]


def test_Container___getitem___03():
    """
    Get slice from container.
    """

    notes = [
        abjad.Note("c'8"),
        abjad.Note("d'8"),
        abjad.Note("e'8"),
        abjad.Note("f'8"),
    ]
    voice = abjad.Voice(notes)

    assert voice[:1] == notes[:1]
    assert voice[:2] == notes[:2]
    assert voice[:3] == notes[:3]
    assert voice[:4] == notes[:4]


def test_Container___getitem___04():
    """
    Bad index raises IndexError.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")

    with pytest.raises(IndexError):
        voice[99]


def test_Container___getitem___05():
    """
    Bad name raises exception.
    """

    score = abjad.Score()

    with pytest.raises(Exception):
        score["Foo"]


def test_Container___getitem___06():
    """
    Duplicate named contexts raise exception.
    """

    score = abjad.Score(
        [
            abjad.Staff(
                [abjad.Voice(name="First_Violin_Voice")],
                name="First_Violin_Staff",
            ),
            abjad.Staff(
                [abjad.Voice(name="Cello_Voice")],
                name="Cello_Staff",
            ),
        ]
    )

    assert score["First_Violin_Voice"].name == "First_Violin_Voice"

    score["Cello_Staff"].append(abjad.Voice(name="First_Violin_Voice"))

    with pytest.raises(Exception):
        score["First_Violin_Voice"]

    extra_first_violin_voice = score["Cello_Staff"].pop()

    assert score["First_Violin_Voice"].name == "First_Violin_Voice"
    assert score["First_Violin_Voice"] is not extra_first_violin_voice


def test_Container___init___01():
    """
    Initialize empty container.
    """

    container = abjad.Container([])

    assert isinstance(container, abjad.Container)
    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
        }
        """
    )


def test_Container___init___02():
    """
    Initialize container with LilyPond note-entry string.
    """

    container = abjad.Container("c'8 d'8 e'8")

    assert isinstance(container, abjad.Container)
    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            d'8
            e'8
        }
        """
    )


def test_Container___setitem___01():
    """
    Replaces in-score leaf with out-of-score leaf.
    """

    voice = abjad.Voice("c'8 [ d'8 ] e'8 f'8")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            f'8
        }
        """
    )

    voice[1] = abjad.Note("c''8")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            c''8
            e'8
            f'8
        }
        """
    )

    assert abjad.wf.wellformed(voice)


def test_Container___setitem___02():
    """
    Replaces in-score leaf with out-of-score container.
    """

    voice = abjad.Voice("c'8 [ d'8 ] e'8 f'8")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            f'8
        }
        """
    )

    voice[1] = abjad.Container("c'16 c'16 c'16")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            {
                c'16
                c'16
                c'16
            }
            e'8
            f'8
        }
        """
    )

    assert abjad.wf.wellformed(voice)


def test_Container___setitem___03():
    """
    Replaces in-score container with out-of-score leaf.
    """

    voice = abjad.Voice("{ c'8 [ d'8 } { e'8 f'8 ] }")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                d'8
            }
            {
                e'8
                f'8
                ]
            }
        }
        """
    )

    voice[1] = abjad.Note("c''8")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                d'8
            }
            c''8
        }
        """
    )

    assert abjad.wf.wellformed(voice)


def test_Container___setitem___04():
    """
    Replaces in-score container with out-of-score tuplet.
    """

    voice = abjad.Voice("{ c'8 d'8 } { e'8 f'8 }")
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                d'8
            }
            {
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    voice[1] = abjad.Tuplet((2, 3), "c'8 d'8 e'8")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                d'8
            }
            \tuplet 3/2
            {
                c'8
                d'8
                e'8
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container___setitem___05():
    """
    Replaces in-score container with out-of-score leaf.
    """

    voice = abjad.Voice("{ c'8 [ d'8 } { e'8 f'8 ] }")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                d'8
            }
            {
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    voice[1] = abjad.Note("c''8")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                d'8
            }
            c''8
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container___setitem___06():
    """
    Replaces in-score container with out-of-score leaf.
    """

    voice = abjad.Voice(
        [abjad.Container("c'8 c'8 c'8 c'8"), abjad.Container("c'8 c'8 c'8 c'8")]
    )
    voice = abjad.Voice("{ c'8 d'8 e'8 f'8 } { g'8 a'8 b'8 c''8 }")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                d'8
                e'8
                f'8
            }
            {
                g'8
                a'8
                b'8
                c''8
            }
        }
        """
    ), print(abjad.lilypond(voice))

    voice[1] = abjad.Rest("r2")

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                d'8
                e'8
                f'8
            }
            r2
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container___setitem___07():
    """
    Replaces note in one score with note from another score.
    """

    notes = [
        abjad.Note("c'8"),
        abjad.Note("d'8"),
        abjad.Note("e'8"),
        abjad.Note("f'8"),
        abjad.Note("g'8"),
        abjad.Note("a'8"),
    ]

    voice_1 = abjad.Voice(notes[:3])
    abjad.beam(voice_1[:])

    assert abjad.lilypond(voice_1) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            e'8
            ]
        }
        """
    ), print(abjad.lilypond(voice_1))

    voice_2 = abjad.Voice(notes[3:])
    abjad.beam(voice_2[:])

    assert abjad.lilypond(voice_2) == abjad.string.normalize(
        r"""
        \new Voice
        {
            f'8
            [
            g'8
            a'8
            ]
        }
        """
    ), print(abjad.lilypond(voice_2))

    voice_1[1] = voice_2[1]

    assert abjad.lilypond(voice_1) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            g'8
            e'8
            ]
        }
        """
    ), print(abjad.lilypond(voice_1))

    assert abjad.wf.wellformed(voice_1)

    assert abjad.lilypond(voice_2) == abjad.string.normalize(
        r"""
        \new Voice
        {
            f'8
            [
            a'8
            ]
        }
        """
    ), print(abjad.lilypond(voice_2))

    assert abjad.wf.wellformed(voice_2)


def test_Container___setitem___08():
    r"""
    Replaces note in one score with container from another score.
    """

    notes = [
        abjad.Note("c'8"),
        abjad.Note("d'8"),
        abjad.Note("e'8"),
        abjad.Note("f'8"),
        abjad.Note("g'8"),
        abjad.Note("a'8"),
        abjad.Note("b'8"),
    ]
    voice_1 = abjad.Voice(notes[:3])
    abjad.beam(voice_1[:])

    assert abjad.lilypond(voice_1) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            e'8
            ]
        }
        """
    ), print(abjad.lilypond(voice_1))

    voice_2 = abjad.Voice(notes[3:])
    abjad.mutate.wrap(voice_2[1:3], abjad.Container())
    leaves = abjad.select.leaves(voice_2)
    leaves = abjad.select.leaves(voice_2[1])
    abjad.slur(leaves)

    assert abjad.lilypond(voice_2) == abjad.string.normalize(
        r"""
        \new Voice
        {
            f'8
            {
                g'8
                (
                a'8
                )
            }
            b'8
        }
        """
    ), print(abjad.lilypond(voice_2))

    voice_1[1] = voice_2[1]

    assert abjad.lilypond(voice_1) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            {
                g'8
                (
                a'8
                )
            }
            e'8
            ]
        }
        """
    ), print(abjad.lilypond(voice_1))

    assert abjad.wf.wellformed(voice_1)

    assert abjad.lilypond(voice_2) == abjad.string.normalize(
        r"""
        \new Voice
        {
            f'8
            b'8
        }
        """
    ), print(abjad.lilypond(voice_2))

    assert abjad.wf.wellformed(voice_2)


def test_Container___setitem___09():
    r"""Sets leaf between unspanned components."""

    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    staff[2:2] = [abjad.Note("g'8")]

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            d'8
            g'8
            e'8
            f'8
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)


def test_Container___setitem___10():
    r"""Sets leaf between spanned compoennts."""

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:])
    note = abjad.Note("g'8")
    voice[2:2] = [note]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            g'8
            e'8
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container___setitem___11():
    r"""
    Sets multiple leaves between spanned components.
    """

    notes = [
        abjad.Note("c'8"),
        abjad.Note("d'8"),
        abjad.Note("e'8"),
        abjad.Note("f'8"),
        abjad.Note("g'8"),
        abjad.Note("a'8"),
    ]

    beginning = notes[:2]
    middle = notes[2:4]
    end = notes[4:]

    voice = abjad.Voice(beginning + end)
    abjad.beam(voice[:])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            g'8
            a'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    voice[2:2] = middle

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            e'8
            f'8
            g'8
            a'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container___setitem___12():
    r"""
    Replaces multiple spanned leaves with with single leaf.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:])
    note = abjad.Note("c''8")
    voice[1:3] = [note]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            c''8
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container___setitem___13():
    """
    Replaces three spanned leaves with three different leaves.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:])
    notes = [abjad.Note("b'8"), abjad.Note("a'8"), abjad.Note("g'8")]
    voice[1:3] = notes

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            b'8
            a'8
            g'8
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container___setitem___14():
    r"""Replaces in-score container with contents of container."""

    staff = abjad.Staff("{ c'8 [ d'8 } { e'8 f'8 ] }")

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            {
                c'8
                [
                d'8
            }
            {
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(staff))

    container = staff[0]
    staff[0:1] = container[:]

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            [
            d'8
            {
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)
    assert len(container) == 0


def test_Container___setitem___15():
    r"""Sets first slice of staff equal to first element of first container in
    staff.
    """

    staff = abjad.Staff("{ c'8 [ d'8 } { e'8 f'8 ] }")

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            {
                c'8
                [
                d'8
            }
            {
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(staff))

    staff[0:0] = staff[0][:1]

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            [
            {
                d'8
            }
            {
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)


def test_Container___setitem___16():
    r"""Sets first slice of staff equal to contents of first container in
    staff.

    Empties first container in staff.

    Leaves empty container in staff.
    """

    staff = abjad.Staff("{ c'8 [ d'8 } { e'8 f'8 ] }")

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            {
                c'8
                [
                d'8
            }
            {
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(staff))

    staff[0:0] = staff[0][:]

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            [
            d'8
            {
            }
            {
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(staff))


def test_Container___setitem___17():
    r"""Set first slice of staff equal to contents of first container in staff;
    empties first container in staff.

    Sets contents of empty first container in staff equal to first component in
    second container in staff.
    """

    staff = abjad.Staff("{ c'8 [ d'8 } { e'8 f'8 ] }")

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            {
                c'8
                [
                d'8
            }
            {
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(staff))

    container = staff[0]
    staff[0:0] = container[:]
    container[0:0] = staff[-1][:1]

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            [
            d'8
            {
                e'8
            }
            {
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.wf.wellformed(staff)


def test_Container___setitem___18():
    r"""
    Extremely small coequal indices indicate first slice in staff.
    """

    voice = abjad.Voice("c'8 [ d'8 e'8 f'8 ]")
    voice[-1000:-1000] = [abjad.Rest("r8")]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            r8
            c'8
            [
            d'8
            e'8
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            r8
            c'8
            [
            d'8
            e'8
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container___setitem___19():
    r"""Extremely large coequal indices indicate last slice in staff."""

    voice = abjad.Voice("c'8 [ d'8 e'8 f'8 ]")
    voice[1000:1000] = [abjad.Rest("r8")]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            e'8
            f'8
            ]
            r8
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container___setitem___20():
    r"""
    You can use setitem to empty the contents of a container.
    """

    staff = abjad.Staff("c'8 d'8 [ e'8 ] f'8")
    inner_container = abjad.Container()
    abjad.mutate.wrap(staff[1:3], inner_container)
    outer_container = abjad.Container()
    abjad.mutate.wrap(inner_container, outer_container)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            {
                {
                    d'8
                    [
                    e'8
                    ]
                }
            }
            f'8
        }
        """
    ), print(abjad.lilypond(staff))

    # sets contents of outer container to nothing
    outer_container[:] = []

    # outer container is empty and remains in score
    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            {
            }
            f'8
        }
        """
    ), print(abjad.lilypond(staff))

    assert abjad.lilypond(inner_container) == abjad.string.normalize(
        r"""
        {
            d'8
            [
            e'8
            ]
        }
        """
    ), print(abjad.lilypond(inner_container))

    # ALTERNATIVE: use del(container)
    staff = abjad.Staff("c'8 d'8 [ e'8 ] f'8")
    inner_container = abjad.Container()
    abjad.mutate.wrap(staff[1:3], inner_container)
    outer_container = abjad.Container()
    abjad.mutate.wrap(inner_container, outer_container)

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            {
                {
                    d'8
                    [
                    e'8
                    ]
                }
            }
            f'8
        }
        """
    ), print(abjad.lilypond(staff))

    # deletes outer container
    del outer_container[:]

    # outer container is empty and remains in score (as before)
    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        {
            c'8
            {
            }
            f'8
        }
        """
    ), print(abjad.lilypond(staff))

    # inner container leaves are still spanned
    assert abjad.lilypond(inner_container) == abjad.string.normalize(
        r"""
        {
            d'8
            [
            e'8
            ]
        }
        """
    ), print(abjad.lilypond(inner_container))


def test_Container_append_01():
    """
    Append container to voice.
    """

    voice = abjad.Voice("c'8 d'8")
    abjad.beam(voice[:])
    voice.append(abjad.Container("e'8 f'8"))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            {
                e'8
                f'8
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container_append_02():
    """
    Append leaf to tuplet.
    """

    tuplet = abjad.Tuplet((2, 3), "c'8 d'8 e'8")
    abjad.Voice([tuplet])
    abjad.beam(tuplet[:])
    tuplet.append(abjad.Note(5, (1, 16)), preserve_duration=True)

    assert abjad.lilypond(tuplet) == abjad.string.normalize(
        r"""
        \tuplet 7/4
        {
            c'8
            [
            d'8
            e'8
            ]
            f'16
        }
        """
    ), print(abjad.lilypond(tuplet))

    assert abjad.wf.wellformed(tuplet)


def test_Container_append_03():
    """
    Trying to append noncomponent to container raises TypeError.
    """

    voice = abjad.Voice("c'8 d'8 e'8")
    abjad.beam(voice[:])

    with pytest.raises(Exception):
        voice.append("foo")
    with pytest.raises(Exception):
        voice.append(99)
    with pytest.raises(Exception):
        voice.append([])
    with pytest.raises(Exception):
        voice.append([abjad.Note(0, (1, 8))])


def test_Container_append_04():
    """
    Append spanned leaf from donor container to recipient container.
    """

    voice_1 = abjad.Voice("c'8 d'8 e'8")
    abjad.beam(voice_1[:])

    assert abjad.lilypond(voice_1) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            e'8
            ]
        }
        """
    ), print(abjad.lilypond(voice_1))

    voice_2 = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice_2[:])

    assert abjad.lilypond(voice_2) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            e'8
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice_2))

    voice_1.append(voice_2[-1])

    assert abjad.lilypond(voice_1) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            e'8
            ]
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice_1))

    assert abjad.lilypond(voice_2) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            e'8
        }
        """
    ), print(abjad.lilypond(voice_2))


def test_Container_append_05():
    """
    Append spanned leaf from donor container to recipient container.
    Donor and recipient containers are the same.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            e'8
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    voice.append(voice[1])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            e'8
            f'8
            ]
            d'8
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container_append_06():
    """
    Can not insert grace container into container.
    """

    staff = abjad.Staff("c' d' e'")
    grace_container = abjad.BeforeGraceContainer("f'16 g'")

    with pytest.raises(Exception):
        staff.append(grace_container)


def test_Container_extend_01():
    """
    Extend container with list of leaves.
    """

    voice = abjad.Voice("c'8 d'8")
    abjad.beam(voice[:])
    voice.extend([abjad.Note("c'8"), abjad.Note("d'8")])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            c'8
            d'8
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container_extend_02():
    """
    Extend container with contents of other container.
    """

    voice_1 = abjad.Voice("c'8 d'8")
    abjad.beam(voice_1[:])

    voice_2 = abjad.Voice("e'8 f'8")
    abjad.beam(voice_2[:])
    voice_1.extend(voice_2)

    assert abjad.lilypond(voice_1) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            [
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice_1))

    assert abjad.wf.wellformed(voice_1)


def test_Container_extend_03():
    """
    Extending container with empty list leaves container unchanged.
    """

    voice = abjad.Voice("c'8 d'8")
    abjad.beam(voice[:])
    voice.extend([])

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container_extend_04():
    """
    Extending one container with empty second container leaves both
    containers unchanged.
    """

    voice = abjad.Voice("c'8 d'8")
    abjad.beam(voice[:])
    voice.extend(abjad.Voice([]))

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)


def test_Container_extend_05():
    """
    Trying to extend container with noncomponent raises TypeError.
    """

    voice = abjad.Voice("c'8 d'8")
    abjad.beam(voice[:])

    with pytest.raises(Exception):
        voice.extend(7)
    with pytest.raises(Exception):
        voice.extend("foo")


def test_Container_extend_06():
    """
    Trying to extend container with noncontainer raises exception.
    """

    voice = abjad.Voice("c'8 d'8")
    abjad.beam(voice[:])

    with pytest.raises(Exception):
        voice.extend(abjad.Note("c'4"))

    with pytest.raises(Exception):
        voice.extend(abjad.Chord("<c' d' e'>4"))


def test_Container_extend_07():
    """
    Extend container with partial and spanned contents of other container.
    """

    voice_1 = abjad.Voice("c'8 d'8")
    abjad.beam(voice_1[:])

    voice_2 = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice_2[:])

    voice_1.extend(voice_2[-2:])

    assert abjad.lilypond(voice_1) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(voice_1))

    assert abjad.lilypond(voice_2) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
        }
        """
    ), print(abjad.lilypond(voice_2))


def test_Container_extend_08():
    """
    Extend container with partial and spanned contents of other container.
    Covered span comes with components from donor container.
    """

    voice_1 = abjad.Voice("c'8 d'8")
    abjad.beam(voice_1[:])

    voice_2 = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice_2[:])
    abjad.slur(voice_2[-2:])

    assert abjad.lilypond(voice_2) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            e'8
            (
            f'8
            )
            ]
        }
        """
    ), print(abjad.lilypond(voice_2))

    voice_1.extend(voice_2[-2:])

    assert abjad.lilypond(voice_1) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            ]
            e'8
            (
            f'8
            )
            ]
        }
        """
    ), print(abjad.lilypond(voice_1))

    assert abjad.lilypond(voice_2) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
        }
        """
    ), print(abjad.lilypond(voice_2))


def test_Container_extend_09():
    """
    Extend container with LilyPond input string.
    """

    container = abjad.Container([])
    container.extend("c'4 ( d'4 e'4 f'4 )")

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'4
            (
            d'4
            e'4
            f'4
            )
        }
        """
    ), print(abjad.lilypond(container))

    assert abjad.wf.wellformed(container)


def test_Container_extend_10():
    """
    Lists must be flattened.
    """

    lists = [
        abjad.makers.make_notes([0, 2], [abjad.Duration(1, 4)]),
        abjad.makers.make_notes([4, 5], [abjad.Duration(1, 4)]),
        abjad.makers.make_notes([7, 9], [abjad.Duration(1, 4)]),
        abjad.makers.make_notes([11, 12], [abjad.Duration(1, 4)]),
    ]
    components = abjad.sequence.flatten(lists, depth=-1)
    container = abjad.Container()
    container.extend(components)

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'4
            d'4
            e'4
            f'4
            g'4
            a'4
            b'4
            c''4
        }
        """
    ), print(abjad.lilypond(container))

    assert abjad.wf.wellformed(container)


def test_Container_index_01():
    """
    Elements that compare equal return different indices in container.
    """

    container = abjad.Container("c'4 c'4 c'4 c'4")

    assert container.index(container[0]) == 0
    assert container.index(container[1]) == 1
    assert container.index(container[2]) == 2
    assert container.index(container[3]) == 3


def test_Container_insert_01():
    """
    Insert component into container at index i.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:])
    voice.insert(0, abjad.Rest((1, 8)))

    assert abjad.wf.wellformed(voice)
    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            r8
            c'8
            [
            d'8
            e'8
            f'8
            ]
        }
        """
    )


def test_Container_insert_02():
    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:])
    voice.insert(1, abjad.Note(1, (1, 8)))

    assert abjad.wf.wellformed(voice)
    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            cs'8
            d'8
            e'8
            f'8
            ]
        }
        """
    )


def test_Container_insert_03():
    voice = abjad.Voice([abjad.Note(n, (1, 8)) for n in range(4)])
    abjad.beam(voice[:])
    voice.insert(4, abjad.Rest((1, 4)))

    assert abjad.wf.wellformed(voice)
    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            cs'8
            d'8
            ef'8
            ]
            r4
        }
        """
    )


def test_Container_insert_04():
    """
    Insert works with really big positive values.
    """

    voice = abjad.Voice([abjad.Note(n, (1, 8)) for n in range(4)])
    abjad.beam(voice[:])
    voice.insert(1000, abjad.Rest((1, 4)))

    assert abjad.wf.wellformed(voice)
    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            cs'8
            d'8
            ef'8
            ]
            r4
        }
        """
    )


def test_Container_insert_05():
    """
    Insert works with negative values.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:])
    voice.insert(-1, abjad.Note(4.5, (1, 8)))

    assert abjad.wf.wellformed(voice)
    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            d'8
            e'8
            eqs'8
            f'8
            ]
        }
        """
    )


def test_Container_insert_06():
    """
    Insert works with really big negative values.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:])
    voice.insert(-1000, abjad.Rest((1, 8)))

    assert abjad.wf.wellformed(voice)
    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            r8
            c'8
            [
            d'8
            e'8
            f'8
            ]
        }
        """
    )


def test_Container_insert_07():
    """
    Inserting a note from one container into another container
    changes note parent from first container to second.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    staff = abjad.Staff("c'8 c'8 c'8 c'8 c'8 c'8 c'8 c'8")
    note = voice[0]
    staff.insert(1, voice[0])

    assert abjad.wf.wellformed(voice)
    assert abjad.wf.wellformed(staff)
    assert note not in voice
    assert note._parent is staff


def test_Container_insert_08():
    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.beam(voice[:])
    voice.insert(1, abjad.Note("cs'8"))

    assert abjad.wf.wellformed(voice)
    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            [
            cs'8
            d'8
            e'8
            f'8
            ]
        }
        """
    )


def test_Container_is_simultaneous_01():
    """
    Is true when container encloses contents in LilyPond << >> brackets, otherwise false.
    """

    assert not abjad.Container().simultaneous
    assert not abjad.Tuplet().simultaneous
    assert abjad.Score().simultaneous
    assert not abjad.Container().simultaneous
    assert not abjad.Staff().simultaneous
    assert abjad.StaffGroup().simultaneous
    assert not abjad.Voice().simultaneous


def test_Container_is_simultaneous_02():
    """
    Is true when container encloses contents in LilyPond << >> brackets, otherwise false.
    """

    container = abjad.Container([])
    container.simultaneous = True
    assert container.simultaneous


def test_Container_is_simultaneous_03():
    """
    ``Container.simultaneous`` is settable.
    """

    container = abjad.Container([])
    assert not container.simultaneous

    container.simultaneous = True
    assert container.simultaneous


def test_Container_is_simultaneous_04():
    """
    Simultaneous container can hold contexts.
    """

    container = abjad.Container([abjad.Voice("c'8 cs'8"), abjad.Voice("d'8 ef'8")])
    container.simultaneous = True

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        <<
            \new Voice
            {
                c'8
                cs'8
            }
            \new Voice
            {
                d'8
                ef'8
            }
        >>
        """
    )


def test_Container_is_simultaneous_05():
    """
    Simultaneous containers must contain only other containers.
    """

    container = abjad.Container(
        [abjad.Container("c'8 c'8 c'8 c'8"), abjad.Container("c'8 c'8 c'8 c'8")]
    )

    container = abjad.Container("c'8 c'8 c'8 c'8")
    with pytest.raises(Exception):
        container.simultaneous = True


def test_Container_pop_01():
    """
    Containers pop leaves correctly.
    Popped leaves abjad.detach from parent.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.slur(voice[:])
    abjad.beam(voice[1:2], beam_lone_notes=True)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            (
            d'8
            [
            ]
            e'8
            f'8
            )
        }
        """
    ), print(abjad.lilypond(voice))

    result = voice.pop(1)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            (
            e'8
            f'8
            )
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)

    "Result is now d'8 [ ]"

    assert abjad.lilypond(result) == "d'8\n[\n]"
    assert abjad.wf.wellformed(result, check_beamed_lone_notes=False)


def test_Container_pop_02():
    """
    Containers pop nested containers correctly.
    """

    voice = abjad.Voice("{ c'8 d'8 } { e'8 f'8 }")
    leaves = abjad.select.leaves(voice)
    abjad.beam(leaves)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                d'8
            }
            {
                e'8
                f'8
                ]
            }
        }
        """
    ), print(abjad.lilypond(voice))

    container = voice.pop()

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                [
                d'8
            }
        }
        """
    ), print(abjad.lilypond(voice))

    assert abjad.wf.wellformed(voice)

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            e'8
            f'8
            ]
        }
        """
    ), print(abjad.lilypond(container))

    assert abjad.wf.wellformed(container)


def test_Container_remove_01():
    """
    Containers remove leaves correctly.
    Leaf abjad.detaches from parentage.
    Leaf returns after removal.
    """

    voice = abjad.Voice("c'8 d'8 e'8 f'8")
    abjad.slur(voice[:])
    abjad.beam(voice[1:2], beam_lone_notes=True)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            (
            d'8
            [
            ]
            e'8
            f'8
            )
        }
        """
    )

    note = voice[1]
    voice.remove(note)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            c'8
            (
            e'8
            f'8
            )
        }
        """
    )

    "Note is now d'8 [ ]"

    assert abjad.lilypond(note) == "d'8\n[\n]"

    assert abjad.wf.wellformed(voice)
    assert abjad.wf.wellformed(note, check_beamed_lone_notes=False)


def test_Container_remove_02():
    """
    Containers remove nested containers correctly.
    Container detaches from parentage.
    Container returns after removal.
    """

    voice = abjad.Voice("{ c'8 d'8 } { e'8 f'8 }")
    container = voice[0]

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                c'8
                d'8
            }
            {
                e'8
                f'8
            }
        }
        """
    )

    voice.remove(container)

    assert abjad.lilypond(voice) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                e'8
                f'8
            }
        }
        """
    )

    assert abjad.wf.wellformed(voice)

    assert abjad.lilypond(container) == abjad.string.normalize(
        r"""
        {
            c'8
            d'8
        }
        """
    )

    assert abjad.wf.wellformed(container)


def test_Container_remove_03():
    """
    Container remove works on identity and not equality.
    """

    note = abjad.Note("c'4")
    container = abjad.Container([abjad.Note("c'4")])

    with pytest.raises(Exception):
        container.remove(note)



from . import _indentlib
from . import contributions as _contributions
from . import enums as _enums
from . import indicators as _indicators
from . import overrides as _overrides


def _get_context_setting_contributions(component, contributions):
    result = []
    if hasattr(component, "_lilypond_type"):
        strings = _overrides.setting(component)._format_in_with_block()
        result.extend(strings)
    else:
        strings = _overrides.setting(component)._format_inline()
        result.extend(strings)
    contributions.context_settings.extend(result)
    contributions.context_settings.sort()


def _get_grob_override_contributions(component, contributions):
    result = []
    once = hasattr(component, "_written_duration")
    grob = _overrides.override(component)
    contributions_ = grob._list_contributions("override", once=once)
    for string in result[:]:
        if "NoteHead" in string and "pitch" in string:
            contributions_.remove(string)
    try:
        written_pitch = component.written_pitch
        arrow = written_pitch.arrow
    except AttributeError:
        arrow = None
    if arrow in (_enums.UP, _enums.DOWN):
        contributions__ = written_pitch._list_contributions()
        contributions_.extend(contributions__)
    contributions.grob_overrides.extend(contributions_)
    contributions.grob_overrides.sort()


def _get_grob_revert_contributions(component, contributions):
    if not hasattr(component, "_written_duration"):
        contributions_ = _overrides.override(component)._list_contributions("revert")
        contributions.grob_reverts.extend(contributions_)
    contributions.grob_reverts.sort()


def _get_indicator_contributions(component, contributions):
    wrappers = []
    for parent in component._get_parentage():
        wrappers_ = parent._get_indicators(unwrap=False)
        wrappers.extend(wrappers_)
    up_markup_wrappers = []
    down_markup_wrappers = []
    neutral_markup_wrappers = []
    context_wrappers = []
    noncontext_wrappers = []
    for wrapper in wrappers:
        if wrapper.annotation:
            continue
        elif not hasattr(wrapper.get_item(), "_get_contributions"):
            continue
        elif (
            wrapper.context is None
            and hasattr(wrapper.get_item(), "format_leaf_children")
            and not getattr(wrapper.get_item(), "format_leaf_children")
            and wrapper.component is not component
        ):
            continue
        elif isinstance(wrapper.unbundle_indicator(), _indicators.Markup):
            if wrapper.direction is _enums.UP:
                up_markup_wrappers.append(wrapper)
            elif wrapper.direction is _enums.DOWN:
                down_markup_wrappers.append(wrapper)
            elif wrapper.direction in (_enums.CENTER, None):
                neutral_markup_wrappers.append(wrapper)
        elif wrapper.context is not None:
            if wrapper.component is component:
                context_wrappers.append(wrapper)
        else:
            noncontext_wrappers.append(wrapper)
    context_wrappers.sort(key=lambda _: type(_.unbundle_indicator()).__name__)
    noncontext_wrappers.sort(key=lambda _: type(_.unbundle_indicator()).__name__)
    for wrappers in (
        up_markup_wrappers,
        down_markup_wrappers,
        neutral_markup_wrappers,
        context_wrappers,
        noncontext_wrappers,
    ):
        for wrapper in wrappers:
            item = wrapper.get_item()
            contributions_ = None
            try:
                contributions_ = item._get_contributions(wrapper=wrapper)
            except TypeError:
                pass
            if contributions_ is None:
                try:
                    contributions_ = item._get_contributions(
                        component=wrapper.component
                    )
                except TypeError:
                    pass
            if contributions_ is None:
                contributions_ = item._get_contributions()
            contributions_.tag_contributions(wrapper.tag, deactivate=wrapper.deactivate)
            if getattr(item, "check_effective_context", False) is True:
                if wrapper._get_effective_context() is None:
                    for list_ in contributions_.get_contribution_lists():
                        list_[:] = [rf"%%% {_} %%%" for _ in list_]
            contributions.update(contributions_)


def _get_contributions_by_site(component) -> _contributions.ContributionsBySite:
    contributions = _contributions.ContributionsBySite()
    _get_indicator_contributions(component, contributions)
    _get_context_setting_contributions(component, contributions)
    _get_grob_override_contributions(component, contributions)
    _get_grob_revert_contributions(component, contributions)
    contributions.freeze_overrides()
    return contributions


def format_component(component) -> str:
    """
    Formats ``component``.
    """
    strings = []
    contributions = _get_contributions_by_site(component)
    strings_ = component._format_absolute_before_site(contributions)
    if strings_:
        strings.append(f"% {_contributions.Sites.ABSOLUTE_BEFORE.name}:")
        strings.extend(strings_)
    strings_ = component._format_before_site(contributions)
    if strings_:
        strings.append(f"% {_contributions.Sites.BEFORE.name}:")
        strings.extend(strings_)
    if hasattr(component, "_format_open_brackets_site"):
        strings_ = component._format_open_brackets_site(contributions)
        if strings_:
            strings.append(f"% {_contributions.Sites.OPEN_BRACKETS.name}:")
            strings.extend(strings_)
    strings_ = component._format_opening_site(contributions)
    if strings_:
        strings.append(f"% {_contributions.Sites.OPENING.name}:")
        strings.extend(strings_)
    strings_ = component._format_contents()
    if strings_:
        strings.extend(strings_)
    strings_ = component._format_closing_site(contributions)
    if strings_:
        strings.append(_indentlib.INDENT + f"% {_contributions.Sites.CLOSING.name}:")
        strings.extend(strings_)
    if hasattr(component, "_format_close_brackets"):
        strings_ = component._format_close_brackets()
        if strings_:
            strings.append(f"% {_contributions.Sites.CLOSE_BRACKETS.name}:")
            strings.extend(strings_)
    strings_ = component._format_after_site(contributions)
    if strings_:
        strings.append(f"% {_contributions.Sites.AFTER.name}:")
        strings.extend(strings_)
    strings_ = component._format_absolute_after_site(contributions)
    if strings_:
        strings.append(f"% {_contributions.Sites.ABSOLUTE_AFTER.name}:")
        strings.extend(strings_)
    assert all(isinstance(_, str) for _ in strings), repr(strings)
    strings = ["" if _.isspace() else _ for _ in strings]
    string = "\n".join(strings)
    return string


def remove_site_comments(string: str) -> str:
    """
    Removes site comments from ``string``.
    """
    site_comments = []
    for site in _contributions.Sites:
        site_comments.append(f"% {site.name}:")
    for type_ in _contributions.Types:
        site_comments.append(f"% {type_.name}:")
    lines = [_ for _ in string.split("\n") if _.strip() not in site_comments]
    string = "\n".join(lines)
    return string



import os
import pathlib
import re
import shutil
import tempfile

from . import contextmanagers as _contextmanagers
from . import illustrators as _illustrators
from . import io as _io
from . import lilypondfile as _lilypondfile
from . import tag as _tag


def as_ly(
    argument,
    ly_file_path,
    *,
    illustrate_function=None,
    tags=False,
    **keywords,
):
    """
    Persists ``argument`` as LilyPond file.

    Returns output path and elapsed formatting time when LilyPond output is written.
    """
    if isinstance(argument, _lilypondfile.LilyPondFile):
        lilypond_file = argument
    elif illustrate_function is not None:
        lilypond_file = illustrate_function(**keywords)
    else:
        lilypond_file = _illustrators.illustrate(argument, **keywords)
    assert ly_file_path is not None, repr(ly_file_path)
    ly_file_path = str(ly_file_path)
    ly_file_path = os.path.expanduser(ly_file_path)
    timer = _contextmanagers.Timer()
    with timer:
        string = lilypond_file._get_lilypond_format()
    if not tags:
        string = _tag.remove_tags(string)
    abjad_formatting_time = timer.elapsed_time
    directory = os.path.dirname(ly_file_path)
    _io._ensure_directory_existence(directory)
    with open(ly_file_path, "w") as file_pointer:
        print(string, file=file_pointer)
    return ly_file_path, abjad_formatting_time


def as_midi(argument, midi_file_path, *, flags: str = "", **keywords):
    """
    Persists ``argument`` as MIDI file.

    Returns 4-tuple of output MIDI path, Abjad formatting time, LilyPond rendering time
    and success boolean.
    """
    if isinstance(argument, _lilypondfile.LilyPondFile) and "score" in argument:
        lilypond_file = argument
    else:
        lilypond_file = _illustrators.illustrate(argument, **keywords)
    assert "score" in lilypond_file, repr(lilypond_file)
    assert midi_file_path is not None, repr(midi_file_path)
    midi_file_path = os.path.expanduser(midi_file_path)
    midi_file_path = pathlib.Path(midi_file_path)
    ly_file_path = midi_file_path.with_suffix(".ly")
    ly_file_path, abjad_formatting_time = as_ly(argument, ly_file_path, **keywords)
    timer = _contextmanagers.Timer()
    with timer:
        success = _io.run_lilypond(ly_file_path, flags=flags)
    lilypond_rendering_time = timer.elapsed_time
    if os.name == "nt":
        extension = "mid"
    else:
        extension = "midi"
    path = os.path.splitext(ly_file_path)[0]
    midi_file_path = f"{path}.{extension}"
    return (
        midi_file_path,
        abjad_formatting_time,
        lilypond_rendering_time,
        success,
    )


def as_pdf(
    argument,
    pdf_file_path,
    *,
    flags: str = "",
    illustrate_function=None,
    tags=False,
    **keywords,
):
    """
    Persists ``argument`` as PDF.

    Returns output path, elapsed formatting time and elapsed rendering time when PDF
    output is written.
    """
    if pdf_file_path is not None:
        pdf_file_path = str(pdf_file_path)
        pdf_file_path = os.path.expanduser(pdf_file_path)
        without_extension = os.path.splitext(pdf_file_path)[0]
        ly_file_path = f"{without_extension}.ly"
    else:
        ly_file_path = None
    result = as_ly(
        argument,
        ly_file_path,
        illustrate_function=illustrate_function,
        tags=tags,
        **keywords,
    )
    ly_file_path, abjad_formatting_time = result
    assert isinstance(ly_file_path, str)
    without_extension = os.path.splitext(ly_file_path)[0]
    pdf_file_path = f"{without_extension}.pdf"
    timer = _contextmanagers.Timer()
    with timer:
        success = _io.run_lilypond(ly_file_path, flags=flags)
    lilypond_rendering_time = timer.elapsed_time
    return (
        pdf_file_path,
        abjad_formatting_time,
        lilypond_rendering_time,
        success,
    )


def as_png(
    argument,
    png_file_path,
    *,
    flags: str = "--png",
    illustrate_function=None,
    preview: bool = False,
    resolution=False,
    tags: bool = False,
    **keywords,
):
    """
    Persists ``argument`` as PNG.

    Autogenerates file path when ``png_file_path`` is none.

    Returns output path(s), elapsed formatting time and elapsed rendering time.
    """
    assert isinstance(flags, str), repr(flags)
    if png_file_path is not None:
        png_file_path = os.path.expanduser(png_file_path)
        without_extension = os.path.splitext(png_file_path)[0]
        ly_file_path = f"{without_extension}.ly"
    else:
        ly_file_path = None
    result = as_ly(
        argument,
        ly_file_path,
        illustrate_function=illustrate_function,
        tags=tags,
        **keywords,
    )
    ly_file_path, abjad_formatting_time = result
    assert isinstance(ly_file_path, str)
    original_directory = os.path.split(ly_file_path)[0]
    original_ly_file_path = ly_file_path
    temporary_directory = tempfile.mkdtemp()
    assert isinstance(temporary_directory, str)
    temporary_ly_file_path = os.path.join(
        temporary_directory, os.path.split(ly_file_path)[1]
    )
    shutil.copy(original_ly_file_path, temporary_ly_file_path)
    if preview:
        flags += " -dpreview"
    if resolution and isinstance(resolution, int):
        flags += f" -dresolution={resolution}"
    timer = _contextmanagers.Timer()
    with timer:
        success = _io.run_lilypond(temporary_ly_file_path, flags=flags)
    lilypond_rendering_time = timer.elapsed_time
    png_file_paths = []
    for file_name in os.listdir(temporary_directory):
        if not file_name.endswith(".png"):
            continue
        source_png_file_path = os.path.join(temporary_directory, file_name)
        target_png_file_path = os.path.join(original_directory, file_name)
        shutil.move(source_png_file_path, target_png_file_path)
        png_file_paths.append(target_png_file_path)
    shutil.rmtree(temporary_directory)
    if 1 < len(png_file_paths):
        _png_page_pattern = re.compile(r".+page(\d+)\.png")

        def _key(path):
            match = _png_page_pattern.match(path)
            assert match is not None
            group = match.groups()[0]
            return group

        png_file_paths.sort(key=_key)
    return (
        tuple(png_file_paths),
        abjad_formatting_time,
        lilypond_rendering_time,
        success,
    )



"""
Classes and functions for modeling spanners: beams, hairpins, slurs, etc.
"""

import typing

from . import _iterlib
from . import bind as _bind
from . import duration as _duration
from . import enums as _enums
from . import indicators as _indicators
from . import iterate as _iterate
from . import parentage as _parentage
from . import score as _score
from . import select as _select
from . import sequence as _sequence
from . import tag as _tag
from . import tweaks as _tweaks


def _apply_tweaks(argument, tweaks, i=None, total=None):
    if not tweaks:
        return argument
    if i is not None:
        assert isinstance(i, int), repr(i)
    if total is not None:
        assert isinstance(total, int), repr(total)
    tweak_objects = []
    for item in tweaks:
        if isinstance(item, _tweaks.Tweak) and item.i is not None:
            item, index = item, item.i
            if 0 <= index and index != i:
                continue
            if index < 0 and index != -(total - i):
                continue
        elif isinstance(item, tuple):
            raise Exception(f"use abjad.Tweak.i instead of tuple: {item}")
        assert isinstance(item, _tweaks.Tweak), repr(item)
        tweak_objects.append(item)
    bundle = _tweaks.bundle(argument, *tweak_objects)
    return bundle


def beam(
    argument: _score.Component | typing.Sequence[_score.Component],
    *,
    beam_lone_notes: bool = False,
    beam_rests: bool | None = True,
    direction: _enums.Vertical | None = None,
    durations: typing.Sequence[_duration.Duration] | None = None,
    span_beam_count: int | None = None,
    start_beam: _indicators.StartBeam | _tweaks.Bundle | None = None,
    stemlet_length: int | float | None = None,
    stop_beam: _indicators.StopBeam | None = None,
    tag: _tag.Tag | None = None,
) -> None:
    r"""
    Attaches beam indicators.

    ..  container:: example

        >>> voice = abjad.Voice("c'8 d' e' f'")
        >>> abjad.beam(voice[:], direction=abjad.UP)
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'8
                ^ [
                d'8
                e'8
                f'8
                ]
            }

        Does not beam rests:

        >>> voice = abjad.Voice("c'8 r e' f'")
        >>> abjad.beam(voice[:], beam_rests=False)
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'8
                r8
                e'8
                [
                f'8
                ]
            }

        Does beam rests:

        >>> voice = abjad.Voice("c'8 r e' f'")
        >>> abjad.beam(voice[:], beam_rests=True)
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'8
                [
                r8
                e'8
                f'8
                ]
            }

        Beams rests and sets stemlet length:

        >>> voice = abjad.Voice("c'8 r e' f'")
        >>> abjad.beam(voice[:], beam_rests=True, stemlet_length=1)
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                \override Staff.Stem.stemlet-length = 1
                c'8
                [
                r8
                e'8
                f'8
                ]
                \revert Staff.Stem.stemlet-length
            }

    """
    original_leaves = list(_iterate.leaves(argument))
    silent_prototype = (_score.MultimeasureRest, _score.Rest, _score.Skip)

    def _is_beamable(argument, beam_rests=False):
        if isinstance(argument, _score.Chord | _score.Note):
            if 0 < argument.written_duration.flag_count:
                return True
        if beam_rests and isinstance(argument, silent_prototype):
            return True
        return False

    leaves = []
    for leaf in original_leaves:
        if not _is_beamable(leaf, beam_rests=beam_rests):
            continue
        leaves.append(leaf)
    runs = []
    run = []
    run.extend(leaves[:1])
    for leaf in leaves[1:]:
        this_index = original_leaves.index(run[-1])
        that_index = original_leaves.index(leaf)
        if this_index + 1 == that_index:
            run.append(leaf)
        else:
            runs.append(run)
            run = [leaf]
    if run:
        runs.append(run)
    runs_ = runs
    if not beam_lone_notes:
        result = _select.nontrivial(runs)
        runs_ = result
    for run in runs_:
        if all(isinstance(_, silent_prototype) for _ in run):
            continue
        start_leaf = run[0]
        stop_leaf = run[-1]
        start_beam_ = start_beam or _indicators.StartBeam()
        stop_beam_ = stop_beam or _indicators.StopBeam()
        _bind.detach(_indicators.StartBeam, start_leaf)
        _bind.attach(start_beam_, start_leaf, direction=direction, tag=tag)
        _bind.detach(_indicators.StopBeam, stop_leaf)
        _bind.attach(stop_beam_, stop_leaf, tag=tag)
        if stemlet_length is None:
            continue
        staff = _parentage.Parentage(start_leaf).get(_score.Staff)
        lilypond_type = getattr(staff, "lilypond_type", "Staff")
        string = rf"\override {lilypond_type}.Stem.stemlet-length = {stemlet_length}"
        literal = _indicators.LilyPondLiteral(string, site="before")
        for indicator in start_leaf._get_indicators():
            if indicator == literal:
                break
        else:
            _bind.attach(literal, start_leaf, tag=tag)
        staff = _parentage.Parentage(stop_leaf).get(_score.Staff)
        lilypond_type = getattr(staff, "lilypond_type", "Staff")
        string = rf"\revert {lilypond_type}.Stem.stemlet-length"
        literal = _indicators.LilyPondLiteral(string, site="after")
        for indicator in stop_leaf._get_indicators():
            if indicator == literal:
                break
        else:
            _bind.attach(literal, stop_leaf, tag=tag)
    if not durations:
        return
    if len(original_leaves) == 1:
        return

    def _leaf_neighbors(leaf, original_leaves):
        assert leaf is not original_leaves[0]
        assert leaf is not original_leaves[-1]
        this_index = original_leaves.index(leaf)
        previous_leaf = original_leaves[this_index - 1]
        previous = 0
        if _is_beamable(previous_leaf, beam_rests=beam_rests):
            previous = previous_leaf.written_duration.flag_count
        next_leaf = original_leaves[this_index + 1]
        next_ = 0
        if _is_beamable(next_leaf, beam_rests=beam_rests):
            next_ = next_leaf.written_duration.flag_count
        return previous, next_

    span_beam_count = span_beam_count or 1
    durations = [_duration.Duration(_) for _ in durations]
    leaf_durations = [_._get_duration() for _ in original_leaves]
    parts = _sequence.partition_by_weights(leaf_durations, durations, overhang=True)
    part_counts = [len(_) for _ in parts]
    parts = _sequence.partition_by_counts(original_leaves, part_counts)
    total_parts = len(parts)
    for i, part in enumerate(parts):
        is_first_part = False
        if i == 0:
            is_first_part = True
        is_last_part = False
        if i == total_parts - 1:
            is_last_part = True
        first_leaf = part[0]
        flag_count = first_leaf.written_duration.flag_count
        if len(part) == 1:
            if not _is_beamable(first_leaf, beam_rests=False):
                continue
            left = flag_count
            right = flag_count
            beam_count = _indicators.BeamCount(left, right)
            _bind.attach(beam_count, first_leaf, tag=tag)
            continue
        if _is_beamable(first_leaf, beam_rests=False):
            if is_first_part:
                left = 0
            else:
                left = span_beam_count
            beam_count = _indicators.BeamCount(left, flag_count)
            _bind.attach(beam_count, first_leaf, tag=tag)
        last_leaf = part[-1]
        if _is_beamable(last_leaf, beam_rests=False):
            flag_count = last_leaf.written_duration.flag_count
            if is_last_part:
                left = flag_count
                right = 0
            else:
                previous, next_ = _leaf_neighbors(last_leaf, original_leaves)
                if previous == next_ == 0:
                    left = right = flag_count
                elif previous == 0:
                    left = 0
                    right = flag_count
                elif next_ == 0:
                    left = flag_count
                    right = 0
                elif previous == flag_count:
                    left = flag_count
                    right = min(span_beam_count, next_)
                elif flag_count == next_:
                    left = min(previous, flag_count)
                    right = flag_count
                else:
                    left = flag_count
                    right = min(previous, flag_count)
            beam_count = _indicators.BeamCount(left, right)
            _bind.attach(beam_count, last_leaf, tag=tag)
        # TODO: eventually remove middle leaf beam counts?
        for middle_leaf in part[1:-1]:
            if not _is_beamable(middle_leaf, beam_rests=beam_rests):
                continue
            if isinstance(middle_leaf, silent_prototype):
                continue
            flag_count = middle_leaf.written_duration.flag_count
            previous, next_ = _leaf_neighbors(middle_leaf, original_leaves)
            if previous == next_ == 0:
                left = right = flag_count
            elif previous == 0:
                left = 0
                right = flag_count
            elif next_ == 0:
                left = flag_count
                right = 0
            elif previous == flag_count:
                left = flag_count
                right = min(flag_count, next_)
            elif flag_count == next_:
                left = min(previous, flag_count)
                right = flag_count
            else:
                left = min(previous, flag_count)
                right = flag_count
            beam_count = _indicators.BeamCount(left, right)
            _bind.attach(beam_count, middle_leaf, tag=tag)


def glissando(
    argument,
    *tweaks,
    allow_repeats: bool = False,
    allow_ties: bool = False,
    hide_middle_note_heads: bool = False,
    hide_middle_stems: bool = False,
    left_broken: bool = False,
    parenthesize_repeats: bool = False,
    right_broken: bool = False,
    right_broken_show_next: bool = False,
    tag: _tag.Tag | None = None,
    zero_padding: bool = False,
):
    r"""
    Attaches glissando indicators.

    ..  container:: example

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8", name="Voice")
        >>> staff = abjad.Staff([voice], name="Staff")
        >>> abjad.glissando(voice[:])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \context Staff = "Staff"
            {
                \context Voice = "Voice"
                {
                    c'8
                    \glissando
                    d'8
                    \glissando
                    e'8
                    \glissando
                    f'8
                }
            }

    ..  container:: example

        Glissando avoids bend-after indicators:

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8", name="Voice")
        >>> staff = abjad.Staff([voice], name="Staff")
        >>> bend_after = abjad.BendAfter()
        >>> abjad.attach(bend_after, voice[1])
        >>> abjad.glissando(voice[:])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \context Staff = "Staff"
            {
                \context Voice = "Voice"
                {
                    c'8
                    \glissando
                    d'8
                    - \bendAfter #'-4
                    e'8
                    \glissando
                    f'8
                }
            }

    ..  container:: example

        Does not allow repeated pitches:

        >>> voice = abjad.Voice("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8", name="Voice")
        >>> staff = abjad.Staff([voice], name="Staff")
        >>> abjad.glissando(voice[:], allow_repeats=True)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \context Staff = "Staff"
            {
                \context Voice = "Voice"
                {
                    a8
                    \glissando
                    a8
                    \glissando
                    b8
                    ~
                    b8
                    \glissando
                    c'8
                    \glissando
                    c'8
                    \glissando
                    d'8
                    ~
                    d'8
                }
            }

    ..  container:: example

        Allows repeated pitches (but not ties):

        >>> voice = abjad.Voice("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8", name="Voice")
        >>> staff = abjad.Staff([voice], name="Staff")
        >>> abjad.glissando(voice[:], allow_repeats=True)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \context Staff = "Staff"
            {
                \context Voice = "Voice"
                {
                    a8
                    \glissando
                    a8
                    \glissando
                    b8
                    ~
                    b8
                    \glissando
                    c'8
                    \glissando
                    c'8
                    \glissando
                    d'8
                    ~
                    d'8
                }
            }

    ..  container:: example

        Allows both repeated pitches and ties:

        >>> voice = abjad.Voice("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8", name="Voice")
        >>> staff = abjad.Staff([voice], name="Staff")
        >>> abjad.glissando(
        ...     voice[:],
        ...     allow_repeats=True,
        ...     allow_ties=True,
        ... )
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \context Staff = "Staff"
            {
                \context Voice = "Voice"
                {
                    a8
                    \glissando
                    a8
                    \glissando
                    b8
                    \glissando
                    ~
                    b8
                    \glissando
                    c'8
                    \glissando
                    c'8
                    \glissando
                    d'8
                    \glissando
                    ~
                    d'8
                }
            }

        Ties are excluded when repeated pitches are not allowed because all ties
        comprise repeated pitches.

    ..  container:: example

        Spans and parenthesizes repeated pitches:

        >>> voice = abjad.Voice("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8", name="Voice")
        >>> staff = abjad.Staff([voice], name="Staff")
        >>> abjad.glissando(
        ...     voice[:],
        ...     allow_repeats=True,
        ...     parenthesize_repeats=True,
        ... )
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \context Staff = "Staff"
            {
                \context Voice = "Voice"
                {
                    a8
                    \glissando
                    \parenthesize
                    a8
                    \glissando
                    b8
                    ~
                    \parenthesize
                    b8
                    \glissando
                    c'8
                    \glissando
                    \parenthesize
                    c'8
                    \glissando
                    d'8
                    ~
                    \parenthesize
                    d'8
                }
            }

    ..  container:: example

        Parenthesizes (but does not span) repeated pitches:

        >>> voice = abjad.Voice("a8 a8 b8 ~ b8 c'8 c'8 d'8 ~ d'8", name="Voice")
        >>> staff = abjad.Staff([voice], name="Staff")
        >>> abjad.glissando(
        ...     voice[:],
        ...     parenthesize_repeats=True,
        ... )
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \context Staff = "Staff"
            {
                \context Voice = "Voice"
                {
                    a8
                    \parenthesize
                    a8
                    \glissando
                    b8
                    ~
                    \parenthesize
                    b8
                    \glissando
                    c'8
                    \parenthesize
                    c'8
                    \glissando
                    d'8
                    ~
                    \parenthesize
                    d'8
                }
            }

    ..  container:: example

        With ``hide_middle_note_heads=True``:

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8", name="Voice")
        >>> staff = abjad.Staff([voice], name="Staff")
        >>> abjad.glissando(voice[:], hide_middle_note_heads=True)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \context Staff = "Staff"
            {
                \context Voice = "Voice"
                {
                    c'8
                    \glissando
                    \hide NoteHead
                    \override Accidental.stencil = ##f
                    \override NoteColumn.glissando-skip = ##t
                    \override NoteHead.no-ledgers = ##t
                    d'8
                    e'8
                    \revert Accidental.stencil
                    \revert NoteColumn.glissando-skip
                    \revert NoteHead.no-ledgers
                    \undo \hide NoteHead
                    f'8
                }
            }

    ..  container:: example

        With ``hide_middle_note_heads=True`` and ``hide_middle_stems=True``:

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8", name="Voice")
        >>> abjad.glissando(
        ...     voice[:],
        ...     hide_middle_note_heads=True,
        ...     hide_middle_stems=True,
        ... )
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \context Voice = "Voice"
            {
                c'8
                \glissando
                \hide NoteHead
                \override Accidental.stencil = ##f
                \override NoteColumn.glissando-skip = ##t
                \override NoteHead.no-ledgers = ##t
                \override Dots.transparent = ##t
                \override Stem.transparent = ##t
                d'8
                e'8
                \revert Accidental.stencil
                \revert NoteColumn.glissando-skip
                \revert NoteHead.no-ledgers
                \undo \hide NoteHead
                \revert Dots.transparent
                \revert Stem.transparent
                f'8
            }

        ..  note:: Respects ``hide_middle_stems`` only when
            ``hide_middle_note_heads=True``.

    ..  container:: example

        With ``right_broken=True``:

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8", name="Voice")
        >>> abjad.glissando(voice[:], right_broken=True)
        >>> abjad.show(voice) # doctest: +SKIP

        LilyPond output looks like this:

        >>> string = abjad.lilypond(voice, tags=True)
        >>> print(string)
        \context Voice = "Voice"
        {
            c'8
            %! abjad.glissando(7)
            \glissando
            d'8
            %! abjad.glissando(7)
            \glissando
            e'8
            %! abjad.glissando(7)
            \glissando
            f'8
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(7)
            %@% \glissando
        }

    ..  container:: example

        With ``right_broken=True`` and ``hide_middle_note_heads=True``:

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8", name="Voice")
        >>> abjad.glissando(
        ...     voice[:],
        ...     right_broken=True,
        ...     hide_middle_note_heads=True,
        ... )
        >>> abjad.show(voice) # doctest: +SKIP

        LilyPond output looks like this:

        >>> string = abjad.lilypond(voice, tags=True)
        >>> print(string)
        \context Voice = "Voice"
        {
            c'8
            %! abjad.glissando(7)
            \glissando
            %! RIGHT_BROKEN
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(0)
            \hide NoteHead
            %! RIGHT_BROKEN
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(0)
            \override Accidental.stencil = ##f
            %! RIGHT_BROKEN
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(0)
            \override NoteColumn.glissando-skip = ##t
            %! RIGHT_BROKEN
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(0)
            \override NoteHead.no-ledgers = ##t
            d'8
            e'8
            %! HIDE_TO_JOIN_BROKEN_SPANNERS
            %! RIGHT_BROKEN
            %! abjad.glissando(4)
            \revert Accidental.stencil
            %! HIDE_TO_JOIN_BROKEN_SPANNERS
            %! RIGHT_BROKEN
            %! abjad.glissando(4)
            \revert NoteColumn.glissando-skip
            %! HIDE_TO_JOIN_BROKEN_SPANNERS
            %! RIGHT_BROKEN
            %! abjad.glissando(4)
            \revert NoteHead.no-ledgers
            %! HIDE_TO_JOIN_BROKEN_SPANNERS
            %! RIGHT_BROKEN
            %! abjad.glissando(4)
            \undo \hide NoteHead
            f'8
        }

    ..  container:: example

        With ``right_broken=True``, ``hide_middle_note_heads=True`` and
        ``right_broken_show_next=True``:

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8", name="Voice")
        >>> abjad.glissando(
        ...     voice[:],
        ...     hide_middle_note_heads=True,
        ...     right_broken=True,
        ...     right_broken_show_next=True,
        ... )
        >>> abjad.show(voice) # doctest: +SKIP

        LilyPond output looks like this:

        >>> string = abjad.lilypond(voice, tags=True)
        >>> print(string)
        \context Voice = "Voice"
        {
            c'8
            %! abjad.glissando(7)
            \glissando
            %! RIGHT_BROKEN
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(0)
            \hide NoteHead
            %! RIGHT_BROKEN
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(0)
            \override Accidental.stencil = ##f
            %! RIGHT_BROKEN
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(0)
            \override NoteColumn.glissando-skip = ##t
            %! RIGHT_BROKEN
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(0)
            \override NoteHead.no-ledgers = ##t
            d'8
            e'8
            %! HIDE_TO_JOIN_BROKEN_SPANNERS
            %! RIGHT_BROKEN
            %! abjad.glissando(4)
            \revert Accidental.stencil
            %! HIDE_TO_JOIN_BROKEN_SPANNERS
            %! RIGHT_BROKEN
            %! abjad.glissando(4)
            \revert NoteColumn.glissando-skip
            %! HIDE_TO_JOIN_BROKEN_SPANNERS
            %! RIGHT_BROKEN
            %! abjad.glissando(4)
            \revert NoteHead.no-ledgers
            %! HIDE_TO_JOIN_BROKEN_SPANNERS
            %! RIGHT_BROKEN
            %! abjad.glissando(4)
            \undo \hide NoteHead
            f'8
            %! RIGHT_BROKEN_SHOW_NEXT
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(5)
            %@% \revert Accidental.stencil
            %! RIGHT_BROKEN_SHOW_NEXT
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(5)
            %@% \revert NoteColumn.glissando-skip
            %! RIGHT_BROKEN_SHOW_NEXT
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(5)
            %@% \revert NoteHead.no-ledgers
            %! RIGHT_BROKEN_SHOW_NEXT
            %! SHOW_TO_JOIN_BROKEN_SPANNERS
            %! abjad.glissando(5)
            %@% \undo \hide NoteHead
        }

    ..  container:: example

        With ``left_broken=True`` (and ``hide_middle_note_heads=True``):

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8", name="Voice")
        >>> abjad.glissando(
        ...     voice[:],
        ...     left_broken=True,
        ...     hide_middle_note_heads=True,
        ... )
        >>> abjad.show(voice) # doctest: +SKIP

        LilyPond output looks like this:

        >>> string = abjad.lilypond(voice, tags=True)
        >>> print(string)
        \context Voice = "Voice"
        {
            %! HIDE_TO_JOIN_BROKEN_SPANNERS
            %! LEFT_BROKEN
            %! abjad.glissando(2)
            \hide NoteHead
            %! HIDE_TO_JOIN_BROKEN_SPANNERS
            %! LEFT_BROKEN
            %! abjad.glissando(2)
            \override Accidental.stencil = ##f
            %! HIDE_TO_JOIN_BROKEN_SPANNERS
            %! LEFT_BROKEN
            %! abjad.glissando(2)
            \override NoteHead.no-ledgers = ##t
            c'8
            %! abjad.glissando(7)
            \glissando
            %! HIDE_TO_JOIN_BROKEN_SPANNERS
            %! LEFT_BROKEN
            %! abjad.glissando(3)
            \override NoteColumn.glissando-skip = ##t
            d'8
            e'8
            %! abjad.glissando(6)
            \revert Accidental.stencil
            %! abjad.glissando(6)
            \revert NoteColumn.glissando-skip
            %! abjad.glissando(6)
            \revert NoteHead.no-ledgers
            %! abjad.glissando(6)
            \undo \hide NoteHead
            f'8
        }

        ..  note:: Respects left-broken only with ``hide_middle_note_heads=True``.

    ..  container:: example

        Tweaks apply to every glissando:

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8", name="Voice")
        >>> abjad.glissando(
        ...     voice[:],
        ...     abjad.Tweak(r"- \tweak style #'trill"),
        ... )
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \context Voice = "Voice"
            {
                c'8
                - \tweak style #'trill
                \glissando
                d'8
                - \tweak style #'trill
                \glissando
                e'8
                - \tweak style #'trill
                \glissando
                f'8
            }

    ..  container:: example

        With ``zero_padding=True`` on fixed pitch:

        >>> voice = abjad.Voice("d'8 d'4. d'4. d'8", name="Voice")

        >>> abjad.glissando(
        ...     voice[:],
        ...     allow_repeats=True,
        ...     zero_padding=True,
        ... )
        >>> for note in voice[1:]:
        ...     abjad.override(note).NoteHead.transparent = True
        ...     abjad.override(note).NoteHead.X_extent = "#'(0 . 0)"
        ...
        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', voice])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \context Voice = "Voice"
            {
                d'8
                - \abjad-zero-padding-glissando
                \glissando
                \once \override NoteHead.X-extent = #'(0 . 0)
                \once \override NoteHead.transparent = ##t
                d'4.
                - \abjad-zero-padding-glissando
                \glissando
                \once \override NoteHead.X-extent = #'(0 . 0)
                \once \override NoteHead.transparent = ##t
                d'4.
                - \abjad-zero-padding-glissando
                \glissando
                \once \override NoteHead.X-extent = #'(0 . 0)
                \once \override NoteHead.transparent = ##t
                d'8
            }

    ..  container:: example

        With ``zero_padding=True`` on changing pitches:

        >>> voice = abjad.Voice("c'8. d'8. e'8. f'8.", name="Voice")
        >>> abjad.glissando(voice[:], zero_padding=True)
        >>> for note in voice[1:-1]:
        ...     abjad.override(note).NoteHead.transparent = True
        ...     abjad.override(note).NoteHead.X_extent = "#'(0 . 0)"
        ...
        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', voice])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \context Voice = "Voice"
            {
                c'8.
                - \abjad-zero-padding-glissando
                \glissando
                \once \override NoteHead.X-extent = #'(0 . 0)
                \once \override NoteHead.transparent = ##t
                d'8.
                - \abjad-zero-padding-glissando
                \glissando
                \once \override NoteHead.X-extent = #'(0 . 0)
                \once \override NoteHead.transparent = ##t
                e'8.
                - \abjad-zero-padding-glissando
                \glissando
                f'8.
            }

    ..  container:: example

        With indexed tweaks:

        >>> voice = abjad.Voice("d'4 d' d' d'", name="Voice")
        >>> abjad.glissando(
        ...     voice[:],
        ...     abjad.Tweak(r"- \tweak color #red", i=0),
        ...     abjad.Tweak(r"- \tweak color #red", i=-1),
        ...     allow_repeats=True,
        ...     zero_padding=True,
        ... )
        >>> for note in voice[1:-1]:
        ...     abjad.override(note).NoteHead.transparent = True
        ...     abjad.override(note).NoteHead.X_extent = "#'(0 . 0)"
        ...
        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', voice])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \context Voice = "Voice"
            {
                d'4
                - \abjad-zero-padding-glissando
                - \tweak color #red
                \glissando
                \once \override NoteHead.X-extent = #'(0 . 0)
                \once \override NoteHead.transparent = ##t
                d'4
                - \abjad-zero-padding-glissando
                \glissando
                \once \override NoteHead.X-extent = #'(0 . 0)
                \once \override NoteHead.transparent = ##t
                d'4
                - \abjad-zero-padding-glissando
                - \tweak color #red
                \glissando
                d'4
            }

    """
    tag = tag or _tag.Tag()

    if right_broken_show_next and not right_broken:
        raise Exception("set right_broken_show_next only when right_broken is true.")

    if hide_middle_stems and not hide_middle_note_heads:
        message = "set hide_middle_stems only when"
        message += " hide_middle_note_heads is true."
        raise Exception(message)

    def _is_last_in_tie_chain(leaf):
        logical_tie = _iterlib._get_logical_tie_leaves(leaf)
        return leaf is logical_tie[-1]

    def _next_leaf_changes_current_pitch(leaf):
        next_leaf = _iterlib._get_leaf(leaf, 1)
        if (
            isinstance(leaf, _score.Note)
            and isinstance(next_leaf, _score.Note)
            and leaf.written_pitch == next_leaf.written_pitch
        ):
            return False
        elif (
            isinstance(leaf, _score.Chord)
            and isinstance(next_leaf, _score.Chord)
            and leaf.written_pitches == next_leaf.written_pitches
        ):
            return False
        return True

    def _parenthesize_leaf(leaf):
        if isinstance(leaf, _score.Note):
            leaf.note_head.is_parenthesized = True
        elif isinstance(leaf, _score.Chord):
            for note_head in leaf.note_heads:
                note_head.is_parenthesized = True

    def _previous_leaf_changes_current_pitch(leaf):
        previous_leaf = _iterlib._get_leaf(leaf, -1)
        if (
            isinstance(leaf, _score.Note)
            and isinstance(previous_leaf, _score.Note)
            and leaf.written_pitch == previous_leaf.written_pitch
        ):
            return False
        elif (
            isinstance(leaf, _score.Chord)
            and isinstance(previous_leaf, _score.Chord)
            and leaf.written_pitches == previous_leaf.written_pitches
        ):
            return False
        return True

    leaves = _select.leaves(argument)
    total = len(leaves) - 1
    for i, leaf in enumerate(leaves):
        if leaf is not leaves[0]:
            if parenthesize_repeats:
                if not _previous_leaf_changes_current_pitch(leaf):
                    _parenthesize_leaf(leaf)
        should_attach_glissando = False
        deactivate_glissando = False
        if leaf._has_indicator(_indicators.BendAfter):
            pass
        elif leaf is leaves[-1]:
            if right_broken is True:
                should_attach_glissando = True
                deactivate_glissando = True
        elif not isinstance(leaf, _score.Chord | _score.Note):
            pass
        elif allow_repeats and allow_ties:
            should_attach_glissando = True
        elif allow_repeats and not allow_ties:
            should_attach_glissando = _is_last_in_tie_chain(leaf)
        elif not allow_repeats and allow_ties:
            if _next_leaf_changes_current_pitch(leaf):
                should_attach_glissando = True
        elif not allow_repeats and not allow_ties:
            if _next_leaf_changes_current_pitch(leaf):
                if _is_last_in_tie_chain(leaf):
                    should_attach_glissando = True
        if hide_middle_note_heads and 3 <= len(leaves):
            if leaf is not leaves[0]:
                should_attach_glissando = False
            if not left_broken and leaf is leaves[1]:
                strings = [
                    r"\hide NoteHead",
                    r"\override Accidental.stencil = ##f",
                    r"\override NoteColumn.glissando-skip = ##t",
                    r"\override NoteHead.no-ledgers = ##t",
                ]
                if hide_middle_stems:
                    strings.extend(
                        [
                            r"\override Dots.transparent = ##t",
                            r"\override Stem.transparent = ##t",
                        ]
                    )
                literal = _indicators.LilyPondLiteral(strings, site="before")
                if right_broken is True:
                    _bind.attach(
                        literal,
                        leaf,
                        tag=tag.append(_tag.Tag("abjad.glissando(0)"))
                        .append(_tag.Tag("SHOW_TO_JOIN_BROKEN_SPANNERS"))
                        .append(_tag.Tag("RIGHT_BROKEN")),
                    )
                else:
                    _bind.attach(
                        literal,
                        leaf,
                        tag=tag.append(_tag.Tag("abjad.glissando(1)")),
                    )
            elif left_broken and leaf is leaves[0]:
                strings = [
                    r"\hide NoteHead",
                    r"\override Accidental.stencil = ##f",
                    r"\override NoteHead.no-ledgers = ##t",
                ]
                if hide_middle_stems:
                    strings.extend(
                        [
                            r"\override Dots.transparent = ##t",
                            r"\override Stem.transparent = ##t",
                        ]
                    )
                literal = _indicators.LilyPondLiteral(strings, site="before")
                _bind.attach(
                    literal,
                    leaf,
                    tag=tag.append(_tag.Tag("abjad.glissando(2)"))
                    .append(_tag.Tag("HIDE_TO_JOIN_BROKEN_SPANNERS"))
                    .append(_tag.Tag("LEFT_BROKEN")),
                )
            elif left_broken and leaf is leaves[1]:
                string = r"\override NoteColumn.glissando-skip = ##t"
                literal = _indicators.LilyPondLiteral(string, site="before")
                _bind.attach(
                    literal,
                    leaf,
                    tag=tag.append(_tag.Tag("abjad.glissando(3)"))
                    .append(_tag.Tag("HIDE_TO_JOIN_BROKEN_SPANNERS"))
                    .append(_tag.Tag("LEFT_BROKEN")),
                )
            if leaf is leaves[-1]:
                strings = [
                    r"\revert Accidental.stencil",
                    r"\revert NoteColumn.glissando-skip",
                    r"\revert NoteHead.no-ledgers",
                    r"\undo \hide NoteHead",
                ]
                if hide_middle_stems:
                    strings.extend(
                        [r"\revert Dots.transparent", r"\revert Stem.transparent"]
                    )
                if right_broken:
                    deactivate_glissando = True
                    literal = _indicators.LilyPondLiteral(strings, site="before")
                    _bind.attach(
                        literal,
                        leaf,
                        deactivate=False,
                        tag=tag.append(_tag.Tag("abjad.glissando(4)"))
                        .append(_tag.Tag("HIDE_TO_JOIN_BROKEN_SPANNERS"))
                        .append(_tag.Tag("RIGHT_BROKEN")),
                    )
                    if right_broken_show_next:
                        literal = _indicators.LilyPondLiteral(strings, site="after")
                        _bind.attach(
                            literal,
                            leaf,
                            deactivate=True,
                            tag=tag.append(_tag.Tag("abjad.glissando(5)"))
                            .append(_tag.Tag("SHOW_TO_JOIN_BROKEN_SPANNERS"))
                            .append(_tag.Tag("RIGHT_BROKEN_SHOW_NEXT")),
                        )
                else:
                    literal = _indicators.LilyPondLiteral(strings, site="before")
                    _bind.attach(
                        literal,
                        leaf,
                        tag=tag.append(_tag.Tag("abjad.glissando(6)")),
                    )
        if should_attach_glissando:
            glissando = _indicators.Glissando(zero_padding=zero_padding)
            glissando = _apply_tweaks(glissando, tweaks, i=i, total=total)
            tag_ = tag.append(_tag.Tag("abjad.glissando(7)"))
            if deactivate_glissando:
                tag_ = tag_.append(_tag.Tag("SHOW_TO_JOIN_BROKEN_SPANNERS"))
            _bind.attach(glissando, leaf, deactivate=deactivate_glissando, tag=tag_)


def hairpin(
    descriptor: str,
    argument: _score.Component | typing.Sequence[_score.Component],
    *,
    direction: _enums.Vertical | None = None,
    tag: _tag.Tag | None = None,
) -> None:
    r"""
    Attaches hairpin indicators.

    ..  container:: example

        With three-part string descriptor:

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> abjad.hairpin("p < f", voice[:], direction=abjad.UP)
        >>> abjad.override(voice[0]).DynamicLineSpanner.staff_padding = 4
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                \once \override DynamicLineSpanner.staff-padding = 4
                c'4
                ^ \p
                ^ \<
                d'4
                e'4
                f'4
                \f
            }

    ..  container:: example

        With two-part string descriptor:

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> abjad.hairpin("< !", voice[:])
        >>> abjad.override(voice[0]).DynamicLineSpanner.staff_padding = 4
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                \once \override DynamicLineSpanner.staff-padding = 4
                c'4
                \<
                d'4
                e'4
                f'4
                \!
            }

    ..  container:: example

        With dynamic objects:

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> start_hairpin = abjad.StartHairpin("o<|")
        >>> bundle = abjad.bundle(start_hairpin, r"- \tweak color #blue")
        >>> stop_dynamic = abjad.Dynamic('"f"')
        >>> abjad.hairpin([bundle, stop_dynamic], voice[:])
        >>> abjad.override(voice[0]).DynamicLineSpanner.staff_padding = 4
        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', voice])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                \once \override DynamicLineSpanner.staff-padding = 4
                c'4
                - \tweak circled-tip ##t
                - \tweak stencil #abjad-flared-hairpin
                - \tweak color #blue
                \<
                d'4
                e'4
                f'4
                _ #(make-dynamic-script
                    (markup
                        #:whiteout
                        #:line (
                            #:general-align Y -2 #:normal-text #:larger "“"
                            #:hspace -0.4
                            #:dynamic "f"
                            #:hspace -0.2
                            #:general-align Y -2 #:normal-text #:larger "”"
                            )
                        )
                    )
            }

    """
    indicators: list = []
    start_dynamic: _indicators.Dynamic | None
    hairpin: _indicators.StartHairpin | None
    stop_dynamic: _indicators.Dynamic | None
    known_shapes = _indicators.StartHairpin("<").known_shapes
    if isinstance(descriptor, str):
        for string in descriptor.split():
            if string in known_shapes:
                hairpin = _indicators.StartHairpin(string)
                indicators.append(hairpin)
            elif string == "!":
                stop_hairpin = _indicators.StopHairpin()
                indicators.append(stop_hairpin)
            else:
                dynamic = _indicators.Dynamic(string)
                indicators.append(dynamic)
    else:
        assert isinstance(descriptor, list), repr(descriptor)
        indicators = descriptor

    start_dynamic, hairpin, stop_dynamic = None, None, None
    if len(indicators) == 1:
        if isinstance(indicators[0], _indicators.Dynamic):
            start_dynamic = indicators[0]
        else:
            hairpin = indicators[0]
    elif len(indicators) == 2:
        if isinstance(indicators[0], _indicators.Dynamic):
            start_dynamic = indicators[0]
            hairpin = indicators[1]
        else:
            hairpin = indicators[0]
            stop_dynamic = indicators[1]
    elif len(indicators) == 3:
        start_dynamic, hairpin, stop_dynamic = indicators
    else:
        raise Exception(indicators)

    if start_dynamic is not None:
        assert isinstance(start_dynamic, _indicators.Dynamic), repr(start_dynamic)

    leaves = _select.leaves(argument)
    start_leaf = leaves[0]
    stop_leaf = leaves[-1]

    if start_dynamic is not None:
        _bind.attach(start_dynamic, start_leaf, direction=direction, tag=tag)
    if hairpin is not None:
        _bind.attach(hairpin, start_leaf, direction=direction, tag=tag)
    if stop_dynamic is not None:
        _bind.attach(stop_dynamic, stop_leaf, tag=tag)


def horizontal_bracket(
    argument: _score.Component | typing.Sequence[_score.Component],
    *,
    start_group: _indicators.StartGroup | _tweaks.Bundle | None = None,
    stop_group: _indicators.StopGroup | None = None,
    tag: _tag.Tag | None = None,
) -> None:
    r"""
    Attaches group indicators.

    ..  container:: example

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> voice.consists_commands.append("Horizontal_bracket_engraver")
        >>> abjad.horizontal_bracket(voice[:])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            \with
            {
                \consists Horizontal_bracket_engraver
            }
            {
                c'4
                \startGroup
                d'4
                e'4
                f'4
                \stopGroup
            }

    ..  container:: example

        Bundle start-group indicators to tweak the padding and outside-staff priority
        of nested analysis brackets:

        >>> voice = abjad.Voice("c'4 d' e' f' c' d' e' f'")
        >>> voice.consists_commands.append("Horizontal_bracket_engraver")
        >>> bundle = abjad.bundle(
        ...     abjad.StartGroup(),
        ...     r"- \tweak color #red",
        ...     r"- \tweak padding 1.5",
        ...     comment="% lexical order 1"
        ... )
        >>> abjad.horizontal_bracket(voice[:4], start_group=bundle)
        >>> bundle = abjad.bundle(
        ...     abjad.StartGroup(),
        ...     r"- \tweak color #red",
        ...     r"- \tweak padding 1.5",
        ...     comment="% lexical order 1"
        ... )
        >>> abjad.horizontal_bracket(voice[4:], start_group=bundle)
        >>> bundle = abjad.bundle(
        ...     abjad.StartGroup(),
        ...     r"- \tweak color #blue",
        ...     r"- \tweak outside-staff-priority 801",
        ...     r"- \tweak padding 1.5",
        ...     comment="% lexical order 0"
        ... )
        >>> abjad.horizontal_bracket(voice[:], start_group=bundle)

        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            \with
            {
                \consists Horizontal_bracket_engraver
            }
            {
                c'4
                % lexical order 0
                - \tweak color #blue
                - \tweak outside-staff-priority 801
                - \tweak padding 1.5
                \startGroup
                % lexical order 1
                - \tweak color #red
                - \tweak padding 1.5
                \startGroup
                d'4
                e'4
                f'4
                \stopGroup
                c'4
                % lexical order 1
                - \tweak color #red
                - \tweak padding 1.5
                \startGroup
                d'4
                e'4
                f'4
                \stopGroup
                \stopGroup
            }

    """
    start_group = start_group or _indicators.StartGroup()
    stop_group = stop_group or _indicators.StopGroup()
    leaves = _select.leaves(argument)
    start_leaf = leaves[0]
    stop_leaf = leaves[-1]
    _bind.attach(start_group, start_leaf, tag=tag)
    _bind.attach(stop_group, stop_leaf, tag=tag)


def ottava(
    argument: _score.Component | typing.Sequence[_score.Component],
    *,
    start_ottava: _indicators.Ottava = _indicators.Ottava(n=1),
    stop_ottava: _indicators.Ottava = _indicators.Ottava(n=0, site="after"),
    tag: _tag.Tag | None = None,
) -> None:
    r"""
    Attaches ottava indicators.

    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.ottava(staff[:])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \ottava 1
                c'4
                d'4
                e'4
                f'4
                \ottava 0
            }

    """
    assert isinstance(start_ottava, _indicators.Ottava), repr(start_ottava)
    assert isinstance(stop_ottava, _indicators.Ottava), repr(stop_ottava)
    leaves = _select.leaves(argument)
    start_leaf = leaves[0]
    stop_leaf = leaves[-1]
    _bind.attach(start_ottava, start_leaf, tag=tag)
    _bind.attach(stop_ottava, stop_leaf, tag=tag)


def phrasing_slur(
    argument: _score.Component | typing.Sequence[_score.Component],
    *,
    direction: _enums.Vertical | None = None,
    start_phrasing_slur: _indicators.StartPhrasingSlur | _tweaks.Bundle | None = None,
    stop_phrasing_slur: _indicators.StopPhrasingSlur | None = None,
    tag: _tag.Tag | None = None,
) -> None:
    r"""
    Attaches phrasing slur indicators.

    ..  container:: example

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> abjad.phrasing_slur(voice[:], direction=abjad.UP)
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                ^ \(
                d'4
                e'4
                f'4
                \)
            }


    """
    start_phrasing_slur = _indicators.StartPhrasingSlur()
    stop_phrasing_slur = _indicators.StopPhrasingSlur()
    leaves = _select.leaves(argument)
    start_leaf = leaves[0]
    stop_leaf = leaves[-1]
    start_phrasing_slur = start_phrasing_slur or _indicators.StartPhrasingSlur()
    stop_phrasing_slur = stop_phrasing_slur or _indicators.StopPhrasingSlur()
    _bind.attach(start_phrasing_slur, start_leaf, direction=direction, tag=tag)
    _bind.attach(stop_phrasing_slur, stop_leaf, tag=tag)


def piano_pedal(
    argument: _score.Component | typing.Sequence[_score.Component],
    *,
    context: str | None = None,
    start_piano_pedal: _indicators.StartPianoPedal | _tweaks.Bundle | None = None,
    stop_piano_pedal: _indicators.StopPianoPedal | None = None,
    tag: _tag.Tag | None = None,
) -> None:
    r"""
    Attaches piano pedal indicators.

    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.piano_pedal(staff[:], context="Staff")
        >>> abjad.setting(staff).pedalSustainStyle = "#'mixed"
        >>> abjad.override(staff).SustainPedalLineSpanner.staff_padding = 5
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override SustainPedalLineSpanner.staff-padding = 5
                pedalSustainStyle = #'mixed
            }
            {
                c'4
                \sustainOn
                d'4
                e'4
                f'4
                \sustainOff
            }

    """
    start_piano_pedal = start_piano_pedal or _indicators.StartPianoPedal()
    stop_piano_pedal = stop_piano_pedal or _indicators.StopPianoPedal()
    leaves = _select.leaves(argument)
    start_leaf = leaves[0]
    stop_leaf = leaves[-1]
    _bind.attach(start_piano_pedal, start_leaf, context=context, tag=tag)
    _bind.attach(stop_piano_pedal, stop_leaf, context=context, tag=tag)


def slur(
    argument: _score.Component | typing.Sequence[_score.Component],
    *,
    direction: _enums.Vertical | None = None,
    start_slur: _indicators.StartSlur | _tweaks.Bundle | None = None,
    stop_slur: _indicators.StopSlur | None = None,
    tag: _tag.Tag | None = None,
) -> None:
    r"""
    Attaches slur indicators.

    ..  container:: example

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> abjad.slur(voice[:], direction=abjad.UP)
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                ^ (
                d'4
                e'4
                f'4
                )
            }


    """
    start_slur = start_slur or _indicators.StartSlur()
    stop_slur = stop_slur or _indicators.StopSlur()
    leaves = _select.leaves(argument)
    start_leaf = leaves[0]
    stop_leaf = leaves[-1]
    _bind.attach(start_slur, start_leaf, direction=direction, tag=tag)
    _bind.attach(stop_slur, stop_leaf, tag=tag)


def text_spanner(
    argument: _score.Component | typing.Sequence[_score.Component],
    *,
    direction: _enums.Vertical | None = None,
    start_text_span: _indicators.StartTextSpan | _tweaks.Bundle | None = None,
    stop_text_span: _indicators.StopTextSpan | None = None,
    tag: _tag.Tag | None = None,
) -> None:
    r"""
    Attaches text span indicators.

    ..  container:: example

        Single spanner:

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> start_text_span = abjad.StartTextSpan(
        ...     left_text=abjad.Markup(r"\upright pont."),
        ...     right_text=abjad.Markup(r"\markup \upright tasto"),
        ...     style=r"\abjad-solid-line-with-arrow",
        ... )
        >>> abjad.text_spanner(
        ...     voice[:], direction=abjad.UP, start_text_span=start_text_span
        ... )
        >>> abjad.override(voice[0]).TextSpanner.staff_padding = 4
        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', voice])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                \once \override TextSpanner.staff-padding = 4
                c'4
                - \abjad-solid-line-with-arrow
                - \tweak bound-details.left.text \markup \concat { \upright pont. \hspace #0.5 }
                - \tweak bound-details.right.text \markup \upright tasto
                ^ \startTextSpan
                d'4
                e'4
                f'4
                \stopTextSpan
            }

    ..  container:: example

        Enchained spanners:

        >>> voice = abjad.Voice("c'4 d' e' f' r")
        >>> start_text_span = abjad.StartTextSpan(
        ...     left_text=abjad.Markup(r"\upright pont."),
        ...     style=r"\abjad-dashed-line-with-arrow",
        ... )
        >>> abjad.text_spanner(voice[:3], start_text_span=start_text_span)
        >>> start_text_span = abjad.StartTextSpan(
        ...     left_text=abjad.Markup(r"\upright tasto"),
        ...     right_text=abjad.Markup(r"\markup \upright pont."),
        ...     style=r"\abjad-dashed-line-with-arrow",
        ... )
        >>> abjad.text_spanner(voice[-3:], start_text_span=start_text_span)
        >>> abjad.override(voice).TextSpanner.staff_padding = 4
        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', voice])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            \with
            {
                \override TextSpanner.staff-padding = 4
            }
            {
                c'4
                - \abjad-dashed-line-with-arrow
                - \tweak bound-details.left.text \markup \concat { \upright pont. \hspace #0.5 }
                \startTextSpan
                d'4
                e'4
                \stopTextSpan
                - \abjad-dashed-line-with-arrow
                - \tweak bound-details.left.text \markup \concat { \upright tasto \hspace #0.5 }
                - \tweak bound-details.right.text \markup \upright pont.
                \startTextSpan
                f'4
                r4
                \stopTextSpan
            }

        >>> voice = abjad.Voice("c'4 d' e' f' r")
        >>> start_text_span = abjad.StartTextSpan(
        ...     left_text=abjad.Markup(r"\upright pont."),
        ...     style=r"\abjad-dashed-line-with-arrow",
        ... )
        >>> abjad.text_spanner(voice[:3], start_text_span=start_text_span)
        >>> start_text_span = abjad.StartTextSpan(
        ...     left_text=abjad.Markup(r"\upright tasto"),
        ...     style=r"\abjad-solid-line-with-hook",
        ... )
        >>> abjad.text_spanner(voice[-3:], start_text_span=start_text_span)
        >>> abjad.override(voice).TextSpanner.staff_padding = 4
        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', voice])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            \with
            {
                \override TextSpanner.staff-padding = 4
            }
            {
                c'4
                - \abjad-dashed-line-with-arrow
                - \tweak bound-details.left.text \markup \concat { \upright pont. \hspace #0.5 }
                \startTextSpan
                d'4
                e'4
                \stopTextSpan
                - \abjad-solid-line-with-hook
                - \tweak bound-details.left.text \markup \concat { \upright tasto \hspace #0.5 }
                \startTextSpan
                f'4
                r4
                \stopTextSpan
            }

    """
    start_text_span = start_text_span or _indicators.StartTextSpan()
    stop_text_span = stop_text_span or _indicators.StopTextSpan()
    leaves = _select.leaves(argument)
    start_leaf = leaves[0]
    stop_leaf = leaves[-1]
    _bind.attach(start_text_span, start_leaf, direction=direction, tag=tag)
    _bind.attach(stop_text_span, stop_leaf, tag=tag)


def tie(
    argument: _score.Component | typing.Sequence[_score.Component],
    *,
    direction: _enums.Vertical | None = None,
    repeat: bool | tuple[int, int] | typing.Callable = False,
    tag: _tag.Tag | None = None,
) -> None:
    r"""
    Attaches tie indicators.

    ..  container:: example

        >>> staff = abjad.Staff("c'4 c' c' c'")
        >>> abjad.tie(staff[:], direction=abjad.UP)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                ^ ~
                c'4
                ^ ~
                c'4
                ^ ~
                c'4
            }

    ..  container:: example

        With repeat ties:

        >>> voice = abjad.Voice("c'4 c' c' c'", name="Voice")
        >>> abjad.tie(voice[:], repeat=True)
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \context Voice = "Voice"
            {
                c'4
                c'4
                \repeatTie
                c'4
                \repeatTie
                c'4
                \repeatTie
            }

    ..  container:: example

        Removes any existing ties before attaching new tie:

        >>> voice = abjad.Voice("c'4 ~ c' ~ c' ~ c'", name="Voice")
        >>> abjad.tie(voice[:])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \context Voice = "Voice"
            {
                c'4
                ~
                c'4
                ~
                c'4
                ~
                c'4
            }

    ..  container:: example

        Ties consecutive chords if all adjacent pairs have at least one pitch
        in common:

        >>> voice = abjad.Voice("<c'>4 <c' d'>4 <d'>4", name="Voice")
        >>> abjad.tie(voice[:])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \context Voice = "Voice"
            {
                <c'>4
                ~
                <c' d'>4
                ~
                <d'>4
            }

    ..  container:: example

        Enharmonics are allowed:

        >>> voice = abjad.Voice("c'4 bs c' dff'", name="Voice")
        >>> abjad.tie(voice[:])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \context Voice = "Voice"
            {
                c'4
                ~
                bs4
                ~
                c'4
                ~
                dff'4
            }

    ..  container:: example

        Repeat tie threshold works like this:

        >>> voice = abjad.Voice("d'4. d'2 d'4. d'2", name="Voice")
        >>> abjad.tie(voice[:], repeat=(4, 8))
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \context Voice = "Voice"
            {
                d'4.
                ~
                d'2
                d'4.
                \repeatTie
                ~
                d'2
            }

    ..  container:: example

        Detaches ties before attach:

        >>> voice = abjad.Voice("d'2 ~ d'8 ~ d'8 ~ d'8 ~ d'8", name="Voice")
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \context Voice = "Voice"
            {
                d'2
                ~
                d'8
                ~
                d'8
                ~
                d'8
                ~
                d'8
            }

        >>> abjad.tie(voice[:], repeat=(4, 8))
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \context Voice = "Voice"
            {
                d'2
                d'8
                \repeatTie
                ~
                d'8
                ~
                d'8
                ~
                d'8
            }

    """
    if callable(repeat):
        pass
    elif repeat in (None, False):

        def inequality(item):
            return item < 0

    elif repeat is True:

        def inequality(item):
            return item >= 0

    else:
        assert isinstance(repeat, tuple) and len(repeat) == 2, repr(repeat)

        def inequality(item):
            return item >= _duration.Duration(repeat)

    leaves = _select.leaves(argument)
    if len(leaves) < 2:
        raise Exception(f"must be two or more notes (not {leaves!r}).")
    for leaf in leaves:
        if not isinstance(leaf, _score.Note | _score.Chord):
            raise Exception(rf"tie note or chord (not {leaf!r}).")
    for current_leaf, next_leaf in _sequence.nwise(leaves):
        duration = current_leaf._get_duration()
        if inequality(duration):
            _bind.detach(_indicators.Tie, current_leaf)
            _bind.detach(_indicators.RepeatTie, next_leaf)
            repeat_tie = _indicators.RepeatTie()
            _bind.attach(repeat_tie, next_leaf, direction=direction, tag=tag)
        else:
            _bind.detach(_indicators.Tie, current_leaf)
            _bind.detach(_indicators.RepeatTie, next_leaf)
            tie = _indicators.Tie()
            _bind.attach(tie, current_leaf, direction=direction, tag=tag)


def trill_spanner(
    argument: _score.Component | typing.Sequence[_score.Component],
    *,
    start_trill_span: _indicators.StartTrillSpan | _tweaks.Bundle | None = None,
    stop_trill_span: _indicators.StopTrillSpan | None = None,
    tag: _tag.Tag | None = None,
) -> None:
    r"""
    Attaches trill spanner indicators.

    ..  container:: example

        >>> voice = abjad.Voice("c'4 d' e' f'")
        >>> abjad.trill_spanner(voice[:])
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                c'4
                \startTrillSpan
                d'4
                e'4
                f'4
                \stopTrillSpan
            }

    """
    start_trill_span = start_trill_span or _indicators.StartTrillSpan()
    stop_trill_span = stop_trill_span or _indicators.StopTrillSpan()
    leaves = _select.leaves(argument)
    start_leaf = leaves[0]
    stop_leaf = leaves[-1]
    _bind.attach(start_trill_span, start_leaf, tag=tag)
    _bind.attach(stop_trill_span, stop_leaf, tag=tag)
